import warnings
import gzip
import bz2
import lzma
import datetime as dt
import numpy as np
from astropy.time import Time, TimeDelta


def parse_str(input_string, metadata):
    """Parse string input.

    Args:
        input_string (str): String to parse.
        metadata (MetaDataSection): Metadata corresponding to this string.

    Returns:
        parsed_str (str): Parsed string.
    """
    return str(input_string)


def parse_datetime(epoch):
    """Convert OEM standard epoch to a DateTime.

    Args:
        epoch (str): OEM epoch string.

    Returns:
        parsed_epoch (DateTime):  Parsed epoch.
    """
    ymd_fmt = "%Y-%m-%d" if epoch.count("-") == 2 else "%Y-%j"
    if "." in epoch:
        return dt.datetime.strptime(
            epoch.replace("Z", "")[:epoch.index(".")+7].strip(),
            f"{ymd_fmt}T%H:%M:%S.%f"
        )
    else:
        return dt.datetime.strptime(
            epoch.replace("Z", "").strip(),
            f"{ymd_fmt}T%H:%M:%S"
        )


def parse_utc(epoch, metadata):
    """Parse OEM standard UTC epoch.

    Args:
        epoch (str): OEM epoch string.

    Returns:
        parsed_epoch (Time): UTC epoch.
    """
    return Time(parse_datetime(epoch), format="datetime", scale="utc")


def parse_epoch(epoch, metadata):
    """Parse OEM standard epoch using metadata TIME_SYSTEM.

    Args:
        epoch (str): OEM epoch string.
        metadata (MetaDataSection): Metadata corresponding to this epoch.

    Returns:
        parsed_epoch (Time): Parsed epoch with assigned time scale. If the
            timescale indicated by TIME_SYSTEM is not supported by astropy,
            then parsed_epoch will warn the user and fall back to DateTime. In
            this case, time calculations may be inaccurate.
    """
    time_system = metadata["TIME_SYSTEM"].lower()
    dt_epoch = parse_datetime(epoch)
    if time_system in Time.SCALES:
        parsed_epoch = Time(dt_epoch, format="datetime", scale=time_system)
    else:
        warnings.warn(
            f"Unsupported TIME_SYSTEM '{time_system}', falling back to "
            f"DateTime. Use caution with time calculations."
        )
        parsed_epoch = dt_epoch
    return parsed_epoch


def _identify_epoch_format(epoch):
    if epoch.count("-") == 2:
        fmt = "isot"
    else:
        fmt = "yday"
    return fmt


def _coerce_epoch_yday(epoch):
    return epoch.replace("-", ":").replace("T", ":")


def _bulk_parse_epochs(epochs, metadata):
    """Parse OEM standard epochs using metadata TIME_SYSTEM.

    Applies time-ordered constraint to input epochs. For faster comparisons,
    the strings are compared directly prior to parsing.

    Args:
        epochs (list of str):
        metadata (MetaDataSection): Metadata corresponding to this epoch.

    Returns:
        parsed_epochs (list of Time):
    """
    time_system = metadata["TIME_SYSTEM"].lower()
    fmt = _identify_epoch_format(epochs[0])
    if fmt != "isot":
        epochs = tuple(_coerce_epoch_yday(epoch) for epoch in epochs)

    if time_system in Time.SCALES:
        parsed_epochs = Time(epochs, format=fmt, scale=time_system)
    else:
        warnings.warn(
            f"Unsupported TIME_SYSTEM '{time_system}', falling back to "
            f"DateTime. Use caution with time calculations."
        )
        parsed_epochs = tuple(parse_epoch(epoch, metadata) for epoch in epochs)

    return parsed_epochs


def parse_integer(input, metadata):
    """Parse integer value.

    Args:
        input: Any input value that can be cast as a number.

    Returns:
        integer (int): Integer equivalent of input.

    Raises:
        ValueError: Invalid integer.
    """
    if float(input).is_integer():
        return int(input)
    else:
        raise ValueError(f"Invalid integer: '{input}'")


def format_float(value):
    """Convert float to a common string format.

    Args:
        value: Any input that can be cast as a float.

    Returns:
        formatted_value (str): Float following standard format.
    """
    return f"{value:+.14e}"


def format_epoch(epoch):
    """Format an epoch in the standard OEM format.

    Args:
        epoch (Time, DateTime): Epoch to convert to string.

    Returns:
        formatted_epoch (str): Epoch in YYYY-MM-DDTHH:MM:SS.sss format.
    """
    return epoch.strftime("%Y-%m-%dT%H:%M:%S.%f")


def require(boolean, message):
    """Require a boolean condition.

    Args:
        boolean (bool): Condition boolean.
        message (str): Error message.

    Raises:
        ValueError: message
    """
    if not boolean:
        raise ValueError(message)


def require_field(field, metadata):
    """Require a field in a dict.

    Args:
        field (str): String containing the required field.
        metadata (dict): Dictionary to check for key.

    Raises:
        ValueError: Missing required header.
    """
    require(field in metadata, f"Missing required header: {field}")


def is_kvn(file_path):
    """Determine if an OEM file is KVN or XML.

    Args:
        file_path (str or Path): Path of file to check.

    returns:
        result (bool): True if file is KVN, false if XML.
    """
    with _open(file_path, "rt") as target_file:
        if "<?xml" in target_file.readline():
            result = False
        else:
            result = True
    return result


def time_range(start_time, stop_time, step_sec):
    """Sample a range of astropy Times.

    Args:
        start_time (Time): Initial time in sample span.
        stop_time (Time): Final time in sample span.
        step_sec (float): Step size in seconds.

    Returns:
        times (generator): Generator of sample astropy Times.
    """
    delta = (stop_time - start_time).sec
    for elapsed in np.arange(0, delta, step_sec):
        yield start_time + TimeDelta(elapsed, format="sec")


def epoch_span_contains(span, epoch):
    """Determine if a given epoch falls within a given timespan.

    Args:
        span (tuple of Time): Pair of Time objects in increasing order.
        epoch (Time): Epoch to compare with span.

    Returns:
        contains (bool): True if input epoch is in the input span, inclusive of
            the endpoint epochs.
    """
    return epoch >= span[0] and epoch <= span[1]


def epoch_span_overlap(span1, span2):
    """Find the overlap between two epoch spans.

    Args:
        span1 (tuple of Time): Range of epochs in increasing order.
        span2 (tuple of Time): Range of epochs in increasing order.

    Returns:
        overlap_range (tuple of Time or None): Overlapping epoch range or None
            if there is no overlap.
    """
    max_start = max(span1[0], span2[0])
    min_end = min(span1[1], span2[1])
    if max_start < min_end:
        overlap_range = (max_start, min_end)
    else:
        overlap_range = None
    return overlap_range


def _get_compression(path):
    headers = {
        b"\x1F\x8b": "gzip",
        b"\x42\x5A\x68": "bz2",
        b"\x5d\x00\x00": "lzma",
        b"\xFD\x37\x7A\x58\x5A\x00": "lzma",
    }
    compression = None
    with open(path, "rb") as fid:
        header = fid.read(6)
        for key, value in headers.items():
            if header.startswith(key):
                compression = value
                break
    return compression


def _open(path, mode, compression=None):
    if mode == "rt":
        compression = _get_compression(path)
    openers = {
        "gzip": gzip.open,
        "bz2": bz2.open,
        "lzma": lzma.open,
        None: open
    }
    return openers[compression](path, mode)



import numpy as np

from oem import CURRENT_VERSION
from oem.tools import require
from oem.base import ConstraintSpecification, Constraint
from oem.compare import StateCompare


COV_XML_ENTRY_MAP = {
    "CX_X": (0, 0),
    "CY_X": (1, 0),
    "CZ_X": (2, 0),
    "CX_DOT_X": (3, 0),
    "CY_DOT_X": (4, 0),
    "CZ_DOT_X": (5, 0),
    "CY_Y": (1, 1),
    "CZ_Y": (2, 1),
    "CX_DOT_Y": (3, 1),
    "CY_DOT_Y": (4, 1),
    "CZ_DOT_Y": (5, 1),
    "CZ_Z": (2, 2),
    "CX_DOT_Z": (3, 2),
    "CY_DOT_Z": (4, 2),
    "CZ_DOT_Z": (5, 2),
    "CX_DOT_X_DOT": (3, 3),
    "CY_DOT_X_DOT": (4, 3),
    "CZ_DOT_X_DOT": (5, 3),
    "CY_DOT_Y_DOT": (4, 4),
    "CZ_DOT_Y_DOT": (5, 4),
    "CZ_DOT_Z_DOT": (5, 5)
}


class ConstrainStateType(Constraint):

    versions = ["1.0"]

    def func(self, state):
        require(
            state.acceleration is None,
            "State in v1.0 OEM cannot have acceleration entries"
        )


class ConstrainStateDimension(Constraint):

    versions = ["*"]

    def func(self, state):
        require(state.position.size == 3, "State position size != 3")
        require(state.velocity.size == 3, "State velocity size != 3")
        if state.acceleration is not None:
            require(
                state.acceleration.size == 3,
                "State acceleration size != 3"
            )


class State(object):
    """Basic Cartesian state.

    Attributes:
        epoch (DateTime): Epoch date and time.
        frame (str): Reference frame.
        center (str): Central body.
        position (ndarray): 3-element array describing the position at epoch.
        velocity (ndarray): 3-element array describing the velocity at epoch.
        acceleration (ndarray): 3-element array describing the acceleration
            at epoch. If unavailable, this attribute is None.
    """

    _constraint_spec = ConstraintSpecification(
        ConstrainStateType,
        ConstrainStateDimension
    )

    def __init__(self, epoch, frame, center, position, velocity,
                 acceleration=None, version=CURRENT_VERSION):
        self.version = version
        self.epoch = epoch
        self.frame = frame
        self.center = center
        self.position = np.array(position)
        self.velocity = np.array(velocity)
        self.acceleration = (
            np.array(acceleration)
            if acceleration is not None else None
        )
        self._constraint_spec.apply(self)

    def __eq__(self, other):
        return (
            self.version == other.version and
            self.epoch == other.epoch and
            (self.position == other.position).all() and
            (self.velocity == other.velocity).all() and
            np.array([self.acceleration == other.acceleration]).all()
        )

    def __sub__(self, other):
        return StateCompare(other, self)

    def __repr__(self):
        return f"State({str(self.epoch)})"

    @classmethod
    def _from_raw_data(cls, data, version, metadata):
        epoch, *state = data
        return cls(
            epoch,
            metadata["REF_FRAME"],
            metadata["CENTER_NAME"],
            state[:3],
            state[3:6],
            acceleration=state[6:] if len(state) == 9 else None,
            version=version
        )

    def copy(self):
        """Create an independent copy of this instance."""
        return State(
            self.epoch,
            self.frame,
            self.center,
            self.position.copy(),
            self.velocity.copy(),
            self.acceleration.copy() if self.has_accel else None,
            version=self.version
        )

    @property
    def has_accel(self):
        return True if self.acceleration is not None else False

    @property
    def vector(self):
        if self.has_accel:
            vec = np.hstack((self.position, self.velocity, self.acceleration))
        else:
            vec = np.hstack((self.position, self.velocity))
        return vec


class Covariance(object):
    """Basic 6x6 covariance.

    Attributes:
        epoch (DateTime): Epoch date and time.
        frame (str): Reference from of this covariance.
        matrix (ndarray): 6x6 covariance matrix.
    """

    def __init__(self, epoch, frame, matrix, version=CURRENT_VERSION):
        self.version = version
        self.epoch = epoch
        self.frame = frame
        self.matrix = np.array(matrix)

    def __eq__(self, other):
        return (
            self.version == other.version and
            self.epoch == other.epoch and
            self.frame == other.frame and
            (self.matrix == other.matrix).all()
        )

    def __repr__(self):
        return f"Covariance({str(self.epoch)})"

    @classmethod
    def _from_raw_data(cls, data, version):
        epoch, frame, *s = data
        matrix = np.array((
            (s[0],  s[1],   s[3],  s[6], s[10], s[15]),
            (s[1],  s[2],   s[4],  s[7], s[11], s[16]),
            (s[3],  s[4],   s[5],  s[8], s[12], s[17]),
            (s[6],  s[7],   s[8],  s[9], s[13], s[18]),
            (s[10], s[11], s[12], s[13], s[14], s[19]),
            (s[15], s[16], s[17], s[18], s[19], s[20]),
        ))
        return cls(epoch, frame, matrix, version=version)

    def copy(self):
        """Create an independent copy of this instance."""
        return Covariance(
            self.epoch,
            self.frame,
            self.matrix.copy(),
            version=self.version
        )



import pytest

from oem import tools


def test_parse_integer():
    tools.parse_integer(1, None)
    tools.parse_integer(1.0, None)
    with pytest.raises(ValueError):
        tools.parse_integer(1.1, None)



import numpy as np

from sgp4.api import Satrec
from astropy.coordinates import (
    CartesianDifferential, CartesianRepresentation, TEME, GCRS
)
from astropy.time import Time
from astropy import units as u
from oem.components import HeaderSection, MetaDataSection, EphemerisSegment
from oem import OrbitEphemerisMessage
from oem.tools import time_range


def _build_header():
    return HeaderSection({
        "CCSDS_OEM_VERS": "2.0",
        "CREATION_DATE": Time.now().isot,
        "ORIGINATOR": "OEM-Python"
    })


def _build_metadata(satrec, start_epoch, stop_epoch):
    return MetaDataSection({
        "OBJECT_NAME": str(satrec.satnum),
        "OBJECT_ID": str(satrec.satnum),
        "CENTER_NAME": "Earth",
        "REF_FRAME": "ICRF",
        "TIME_SYSTEM": "UTC",
        "START_TIME": start_epoch.isot,
        "STOP_TIME": stop_epoch.isot,
    })


def _build_segment(satrec, start_epoch, stop_epoch, step, frame):
    epoch_range = list(time_range(start_epoch, stop_epoch, step))
    position, velocity = _sample_tle_at_epoch_array(satrec, epoch_range, frame)
    return EphemerisSegment(
        _build_metadata(satrec, start_epoch, stop_epoch),
        (epoch_range, *zip(*position), *zip(*velocity))
    )


def _build_oem(satrec, start_epoch, stop_epoch, step, frame):
    return OrbitEphemerisMessage(
        _build_header(),
        [_build_segment(satrec, start_epoch, stop_epoch, step, frame)]
    )


def _sample_tle_at_epoch_array(satrec, epochs, frame):
    if frame not in ("TEME", "ICRF"):
        raise ValueError(f"Unsupported frame: {frame}")
    jd1 = np.array([epoch.jd1 for epoch in epochs])
    jd2 = np.array([epoch.jd2 for epoch in epochs])
    err, r, v = satrec.sgp4_array(jd1, jd2)
    if any(err):
        raise ValueError("SGP4 propagation failed!")
    else:
        teme_p = CartesianRepresentation(r.T*u.km)
        teme_v = CartesianDifferential(v.T*u.km/u.s)
        states = TEME(teme_p.with_differentials(teme_v), obstime=epochs)
        if frame == "ICRF":
            states = states.transform_to(GCRS(obstime=epochs))
        return (
            states.cartesian.get_xyz().value.T,
            states.cartesian.differentials["s"].get_d_xyz().value.T
        )


def tle_to_oem(tle, start_epoch, stop_epoch, step, frame="ICRF"):
    """Create an OEM instance from a Two-Line Element set.

    Args:
        tle (list of str): List of two or three line element strings.
        start_epoch (Time): Output OEM start time.
        stop_epoch (Time): Output OEM stop time.
        step (float): Output OEM step time in seconds.
        frame (str, optional): Desired output frame. Currently supported
            options are "ICRF" and "TEME". Default is "ICRF".

    Returns:
        oem (OrbitEphemerisMessage): Converted OEM instance.

    Raises:
        ValueError: Unsupported frame.
    """
    satrec = Satrec.twoline2rv(*tle[-2:])
    return satrec_to_oem(satrec, start_epoch, stop_epoch, step, frame=frame)


def satrec_to_oem(satrec, start_epoch, stop_epoch, step, frame="ICRF"):
    """Create an OEM instance from an sgp4.api.Satrec instance.

    Args:
        satrec (Satrec): Satrec instance containing desired TLE.
        start_epoch (Time): Output OEM start time.
        stop_epoch (Time): Output OEM stop time.
        step (float): Output OEM step time in seconds.
        frame (str, optional): Desired output frame. Currently supported
            options are "ICRF" and "TEME". Default is "ICRF".

    Returns:
        oem (OrbitEphemerisMessage): Converted OEM instance.

    Raises:
        ValueError: Unsupported frame.
    """
    return _build_oem(satrec, start_epoch, stop_epoch, step, frame)



from lxml.etree import SubElement

from oem import CURRENT_VERSION
from oem.tools import (
    parse_epoch, parse_integer, parse_str, format_epoch, require,
    require_field)
from oem.base import (
    KeyValueSection, HeaderField, ConstraintSpecification, Constraint)


class ConstrainMetaDataTime(Constraint):
    '''Apply constraints to metadata START_TIME and STOP_TIME'''

    versions = ["*"]

    def func(self, metadata):
        require_field("START_TIME", metadata)
        require_field("STOP_TIME", metadata)
        require(
            metadata["START_TIME"] <= metadata["STOP_TIME"],
            "START_TIME is before STOP_TIME"
        )


class ConstrainMetadataUseableTime(Constraint):
    '''Apply constraints to USEABLE_START_TIME & USEABLE_STOP_TIME'''

    versions = ["*"]

    def func(self, metadata):
        if "USEABLE_START_TIME" in metadata or "USEABLE_STOP_TIME" in metadata:
            require(
                "USEABLE_START_TIME" in metadata,
                "USEABLE_STOP_TIME provided without USEABLE_START_TIME"
            )
            require(
                "USEABLE_STOP_TIME" in metadata,
                "USEABLE_START_TIME provided without USEABLE_STOP_TIME"
            )
            require(
                metadata["USEABLE_START_TIME"]
                <= metadata["USEABLE_STOP_TIME"],
                "USEABLE_START_TIME after USEABLE_STOP_TIME"
            )
            require(
                metadata["USEABLE_START_TIME"] >= metadata["START_TIME"],
                "USEABLE_START_TIME before START_TIME"
            )
            require(
                metadata["USEABLE_STOP_TIME"] <= metadata["STOP_TIME"],
                "USEABLE_STOP_TIME after STOP_TIME"
            )


class ConstrainMetaDataInterpolation(Constraint):
    '''Apply constraints to metadata INTERPOLATION and INTERPOLATION_DEGREE'''

    versions = ["*"]

    def func(self, metadata):
        if "INTERPOLATION" in metadata:
            require_field("INTERPOLATION_DEGREE", metadata)
            require(
                float(metadata["INTERPOLATION_DEGREE"]).is_integer(),
                "Interpolation degree is not an integer"
            )


class ConstrainMetaDataRefFrameEpoch(Constraint):
    '''Apply constraints to metadata REF_FRAME_EPOCH'''

    versions = ["1.0"]

    def func(self, metadata):
        require(
            "REF_FRAME_EPOCH" not in metadata,
            "Metadata keyword 'REF_FRAME_EPOCH' not supported in OEM v1.0"
        )


class ConstrainMetaDataMessageId(Constraint):
    '''Apply constraints to metadata MESSAGE_ID'''

    versions = ["1.0", "2.0"]

    def func(self, metadata):
        require(
            "MESSAGE_ID" not in metadata,
            "Metadata keyword 'MESSAGE_ID' not supported in OEM v1.0 and v2.0"
        )


class MetaDataSection(KeyValueSection):
    """OEM metadata section.

    Container for a single OEM metadata section.

    Examples:
        This class behaves similar to a dict allowing membership checks,
        iteration over keys, and value set/get.

        >>> "OBJECT_NAME" in metadata:
        True

        >>> keys = [key for key in metadata]

        >>> metadata["CENTER_NAME"] = 'Mars'

        >>> metadata["CENTER_NAME"]
        'Mars'
    """

    _field_spec = {
        "OBJECT_NAME": HeaderField(parse_str, str, required=True),
        "OBJECT_ID": HeaderField(parse_str, str, required=True),
        "CENTER_NAME": HeaderField(parse_str, str, required=True),
        "REF_FRAME": HeaderField(parse_str, str, required=True),
        "TIME_SYSTEM": HeaderField(parse_str, str, required=True),
        "START_TIME": HeaderField(parse_epoch, format_epoch, required=True),
        "STOP_TIME": HeaderField(parse_epoch, format_epoch, required=True),
        "REF_FRAME_EPOCH": HeaderField(parse_epoch, format_epoch),
        "USEABLE_START_TIME": HeaderField(parse_epoch, format_epoch),
        "USEABLE_STOP_TIME": HeaderField(parse_epoch, format_epoch),
        "INTERPOLATION": HeaderField(parse_str, str),
        "INTERPOLATION_DEGREE": HeaderField(parse_integer, str),
        "MESSAGE_ID": HeaderField(parse_str, str),
    }
    _constraint_spec = ConstraintSpecification(
        ConstrainMetaDataTime,
        ConstrainMetadataUseableTime,
        ConstrainMetaDataInterpolation,
        ConstrainMetaDataRefFrameEpoch,
        ConstrainMetaDataMessageId,
    )

    def __init__(self, metadata, version=CURRENT_VERSION):
        self.version = version
        self._parse_fields(metadata)
        self._constraint_spec.apply(self)

    def __eq__(self, other):
        return (
            self.version == other.version and
            self._fields.keys() == other._fields.keys() and
            all(self[key] == other[key] for key in self)
        )

    def __repr__(self):
        start = str(self.useable_start_time)
        stop = str(self.useable_stop_time)
        return f"MetaDataSection({start}, {stop})"

    @classmethod
    def _from_raw_data(cls, segment, version):
        return cls(segment, version=version)

    def _to_string(self):
        lines = "META_START\n"
        lines += "\n".join(self._format_fields()) + "\n"
        lines += "META_STOP\n"
        return lines

    def _to_xml(self, parent):
        for key, value in self._fields.items():
            SubElement(parent, key).text = value

    def copy(self):
        """Create an independent copy of this instance."""
        return MetaDataSection(self._fields.copy(), version=self.version)

    @property
    def useable_start_time(self):
        """Return epoch of start of useable state data range"""
        return (
            self["USEABLE_START_TIME"]
            if "USEABLE_START_TIME" in self
            else self["START_TIME"]
        )

    @property
    def useable_stop_time(self):
        """Return epoch of end of useable state data range"""
        return (
            self["USEABLE_STOP_TIME"]
            if "USEABLE_STOP_TIME" in self
            else self["STOP_TIME"]
        )



"""Test parsing sample OEMS.
"""
import pytest
import glob
import tempfile
from pathlib import Path

from oem import OrbitEphemerisMessage
from oem.tools import is_kvn


SAMPLE_DIR = Path(__file__).parent / "samples"


def _get_test_files(version="*", validity="*"):
    samples = SAMPLE_DIR / version / validity / "*.oem*"
    return sorted([entry for entry in glob.glob(str(samples))])


@pytest.mark.parametrize("file_path", _get_test_files(validity="valid"))
def test_valid_samples(file_path):
    oem = OrbitEphemerisMessage.open(file_path)
    assert oem.span[0] <= oem.span[1]

    for segment in oem:
        if not segment.has_accel:
            for state in segment.states:
                assert state.acceleration is None
        for covariance in segment.covariances:
            assert covariance.matrix.shape == (6, 6)

        assert segment.useable_start_time in segment
        assert segment.useable_stop_time in segment
        assert len(oem.states) > 0
        assert len(oem.covariances) >= 0

    with tempfile.TemporaryDirectory() as tmp_dir:
        written_oem_path = Path(tmp_dir) / "written.oem"
        fmt = "xml" if is_kvn(file_path) else "kvn"
        OrbitEphemerisMessage.convert(file_path, written_oem_path, fmt)
        written_oem = OrbitEphemerisMessage.open(written_oem_path)
        assert written_oem == oem


@pytest.mark.parametrize("file_path", _get_test_files(validity="invalid"))
def test_invalid_samples(file_path):
    with pytest.raises(Exception):
        OrbitEphemerisMessage.open(file_path)


@pytest.mark.parametrize("file_path", _get_test_files(validity="valid"))
def test_convert(file_path):
    with tempfile.TemporaryDirectory() as tmp_dir:
        converted_xml_path = Path(tmp_dir) / "written.oem"
        OrbitEphemerisMessage.convert(
            file_path,
            converted_xml_path,
            "kvn"
        )
        converted_xml = OrbitEphemerisMessage.open(converted_xml_path)

        converted_kvn_path = Path(tmp_dir) / "written.oem"
        OrbitEphemerisMessage.convert(
            converted_xml_path,
            converted_kvn_path,
            "xml"
        )
        converted_kvn = OrbitEphemerisMessage.open(converted_kvn_path)

        assert converted_xml == converted_kvn


def test_compare():
    test_files = _get_test_files("v2_0", "valid")
    oem1 = OrbitEphemerisMessage.open(test_files[0])
    oem2 = OrbitEphemerisMessage.open(test_files[1])
    assert oem1 == oem1
    assert oem2 == oem2
    assert oem1 != oem2


@pytest.mark.parametrize("file_path", _get_test_files(validity="valid"))
def test_copy(file_path):
    oem1 = OrbitEphemerisMessage.open(file_path)
    oem2 = oem1.copy()
    assert oem1 is not oem2 and oem1 == oem2


@pytest.mark.parametrize("compression", ("gzip", "bz2", "lzma"))
def test_compression(compression):
    file_path = _get_test_files(validity="valid")[0]
    oem = OrbitEphemerisMessage.open(file_path)
    with tempfile.TemporaryDirectory() as tmp_dir:
        written_oem_path = Path(tmp_dir) / "written.oem"
        oem.save_as(written_oem_path, compression=compression)
        oem_readback = OrbitEphemerisMessage.open(written_oem_path)
        assert oem == oem_readback



from lxml.etree import ElementTree, Element, SubElement

from oem import components
from oem.base import Constraint, ConstraintSpecification
from oem.tools import require, is_kvn, _open
from oem.compare import EphemerisCompare
from oem.parsers import parse_kvn_oem, parse_xml_oem


class ConstrainOemTimeSystem(Constraint):
    """Apply constraints to OEM TIME_SYSTEM."""

    versions = ["*"]

    def func(self, oem):
        time_system = None
        for segment in oem:
            if time_system is None:
                time_system = segment.metadata["TIME_SYSTEM"]
            else:
                require(
                    segment.metadata["TIME_SYSTEM"] == time_system,
                    "TIME_SYSTEM not fixed in OEM"
                )


class ConstrainOemObject(Constraint):
    """Apply constraint to OEM OBJECT_NAME and OBJECT_ID."""

    versions = ["*"]

    def func(self, oem):
        object_name = oem._segments[0].metadata["OBJECT_NAME"]
        object_id = oem._segments[0].metadata["OBJECT_ID"]
        for segment in oem:
            require(
                segment.metadata["OBJECT_NAME"] == object_name,
                "OBJECT_NAME not fixed in OEM"
            )
            require(
                segment.metadata["OBJECT_ID"] == object_id,
                "OBJECT_ID not fixed in OEM"
            )


class ConstrainOemStates(Constraint):
    '''Apply constraints to OEM data sections'''

    versions = ["*"]

    def func(self, oem):
        if oem.version == "1.0":
            self.v1_0(oem)
        else:
            self.v2_0(oem)

    def v1_0(self, oem):
        require(
            all(
                (oem._segments[idx].metadata["STOP_TIME"]
                 <= oem._segments[idx+1].metadata["START_TIME"])
                for idx in range(len(oem._segments)-1)
            ),
            "Data section state epochs overlap"
        )

    def v2_0(self, oem):
        require(
            all(
                (oem._segments[idx].useable_stop_time
                 <= oem._segments[idx+1].useable_start_time)
                for idx in range(len(oem._segments)-1)
            ),
            "Data section state epochs overlap"
        )


class OrbitEphemerisMessage(object):
    """Python representation of an Orbit Ephemeris Message.

    This class provides the primary interface between the OEM module and an
    OEM file.

    Attributes:
        header (HeaderSection): Object containing the OEM header section.

    Examples:
        The `OrbitEphemerisMessage` class can load directly from a file:

        >>> ephemeris = OrbitEphemerisMessage.open(file_path)

        An OEM is made up of one or more data segments available through an
        iterator:

        >>> for segment in ephemeris:
        ...     for state in segment:
        ...         # Iterate through states
        ...         pass
        ...     for covariance in segment.covariances:
        ...         # Iterate through covariances
        ...         pass

        It is also possible to iterate through the states and covariances in
        all segments with the `.states` and `.covariances` properties.

        To determine if a particular epoch is contained in the useable time
        range of any of the segments in an ephemeris, use `in`:

        >>> epoch in ephemeris
        True

        To sample a state at an arbitrary epoch, simply call the ephemeris with
        an astropy Time object

        >>> epoch = Time("2020-01-01T00:00:00", scale="utc")
        >>> ephemeris(epoch)
        State(2020-01-01T00:00:00.000)

        Note that this type of sampling is only supported if the time system of
        the target ephemeris is supported by astropy Time objects.

        The `save_as` method enables saving of copies of an OEM in both KVN and
        XML formats.

        >>> oem.save_as("new.oem", file_format="xml")

        To convert directly between KVN and XML formats, use the `convert`
        class method. For example, to convert a KVN OEM to XML:

        >>> oem.convert("input.oem", "output.oem", "xml")
    """

    _constraint_spec = ConstraintSpecification(
        ConstrainOemTimeSystem,
        ConstrainOemObject,
        ConstrainOemStates
    )

    def __init__(self, header, segments):
        '''Create an Orbit Ephemeris Message.

        Args:
            header (HeaderSection): Object containing the OEM header section.
            segments (list): List of OEM EphemerisSegments.
        '''
        self.header = header
        self.version = self.header["CCSDS_OEM_VERS"]
        self._segments = segments
        self._constraint_spec.apply(self)

    def __call__(self, epoch):
        for segment in self:
            if epoch in segment:
                return segment(epoch)
        else:
            raise ValueError(f"Epoch {epoch} not contained in this ephemeris.")

    def __iter__(self):
        return iter(self._segments)

    def __contains__(self, epoch):
        return any(epoch in segment for segment in self._segments)

    def __eq__(self, other):
        return (
            self.version == other.version and
            self.header == other.header and
            len(self._segments) == len(other._segments) and
            all(
                this_segment == other_segment
                for this_segment, other_segment
                in zip(self._segments, other._segments)
            )
        )

    def __sub__(self, other):
        return EphemerisCompare(other, self)

    def __repr__(self):
        return f"OrbitEphemerisMessage(v{self.version})"

    @classmethod
    def _from_kvn_oem(cls, file_path):
        with _open(file_path, "rt") as ephem_file:
            return cls._from_raw_data(parse_kvn_oem(ephem_file))

    @classmethod
    def _from_xml_oem(cls, file_path):
        with _open(file_path, "rt") as ephem_file:
            return cls._from_raw_data(parse_xml_oem(ephem_file))

    @classmethod
    def _from_raw_data(cls, data):
        raw_header, raw_segments = data
        header = components.HeaderSection._from_raw_data(raw_header)
        segments = [
            components.EphemerisSegment._from_raw_data(
                raw_segment, header.version
            ) for raw_segment in raw_segments
        ]
        return cls(header, segments)

    @classmethod
    def open(cls, file_path):
        """Open an Orbit Ephemeris Message file.

        This method supports both KVN and XML formats.

        Args:
            file_path (str or Path): Path of file to read.

        Returns:
            OrbitEphemerisMessage: New OEM instance.
        """
        if is_kvn(file_path):
            oem = cls._from_kvn_oem(file_path)
        else:
            oem = cls._from_xml_oem(file_path)
        return oem

    @classmethod
    def convert(cls, in_file_path, out_file_path, file_format):
        """Convert an OEM to a particular file format.

        This method will succeed and produce an output file even if the input
        file is already in the desired format. Comments are not preserved.

        Args:
            in_file_path (str or Path): Path to original ephemeris.
            out_file_path (str or Path): Desired path for converted ephemeris.
            file_format (str): Desired output format. Options are
                'kvn' and 'xml'.
        """
        cls.open(in_file_path).save_as(out_file_path, file_format=file_format)

    def copy(self):
        """Create an independent copy of this instance."""
        return OrbitEphemerisMessage(
            self.header.copy(),
            [segment.copy() for segment in self]
        )

    def steps(self, step_size):
        """Sample Ephemeris at equal time intervals.

        This method returns a generator producing states at equal time
        intervals spanning the useable duration of all segments in the
        parent OEM.

        Args:
            step_size (float): Sample step size in seconds.

        Yields:
            State: Sample state.

        Examples:
            Sample states at 60-second intervals:

            >>> for state in oem.steps(60):
            ...     pass

            Note that spacing between steps will only be constant within
            segments; when crossing from one segment to another the spacing
            will vary. To avoid this behavior with multi-segment OEMs, use the
            segment interface directly:

            >>> for segment in oem:
            ...    for state in segment.steps(60):
            ...        pass
        """
        for segment in self:
            for state in segment.steps(step_size):
                yield state

    def resample(self, step_size, in_place=False):
        """Resample ephemeris data.

        Replaces the existing ephemeris state data in this OEM with new states
        sampled at the desired sampling interval. The new sampling applies to
        all segments contained in this OEM.

        Args:
            step_size (float): Sample step size in seconds.
            in_place (bool, optional): Toggle in-place resampling. Default
                is False.

        Returns:
            OrbitEphemerisMessage: Resampled OEM. Output is an indepedent
                instance if in_place is True.

        Examples:
            Open an ephemeris file, convert it to a 60-second sampling interval
            and save the result to a new file:

            >>> oem = OrbitEphemerisMessage.open("input.oem")
            >>> oem.resample(60, in_place=True)
            >>> oem.save_as("output.oem")

            To do the same thing without in-place operations:

            >>> oem = OrbitEphemerisMessage.open("input.oem")
            >>> new_oem = oem.resample(60)
            >>> new_oem.save_as("output.oem")
        """
        if in_place:
            for segment in self:
                segment.resample(step_size, in_place=True)
        else:
            oem = self.copy().resample(step_size, in_place=True)
        return oem if not in_place else self

    def save_as(self, file_path, file_format="kvn", compression=None):
        """Write OEM to file.

        Args:
            file_path (str or Path): Desired path for output ephemeris.
            file_format (str, optional): Type of file to output. Options are
                'kvn' and 'xml'. Default is 'kvn'.
        """
        with _open(file_path, "wb", compression) as output_file:
            if file_format == "kvn":
                output_file.write(bytes(self._to_kvn_oem(), "utf-8"))
            elif file_format == "xml":
                self._to_xml_oem().write(
                    output_file,
                    pretty_print=True,
                    encoding="utf-8",
                    xml_declaration=True
                )
            else:
                raise ValueError(f"Unrecognized file type: '{file_format}'")

    def _to_kvn_oem(self):
        lines = self.header._to_string() + "\n"
        lines += "".join(entry._to_string() for entry in self._segments)
        return lines

    def _to_xml_oem(self):
        oem = Element("oem", id="CCSDS_OEM_VERS", version=self.version)
        self.header._to_xml(SubElement(oem, "header"))
        body = SubElement(oem, "body")
        for entry in self._segments:
            entry._to_xml(SubElement(body, "segment"))
        return ElementTree(oem)

    @property
    def states(self):
        '''Return a list of states in all segments.'''
        return [
            state
            for segment in self
            for state in segment.states
        ]

    @property
    def covariances(self):
        '''Return a list of covariances in all segments.'''
        return [
            covariance
            for segment in self
            for covariance in segment.covariances
        ]

    @property
    def segments(self):
        return self._segments

    @property
    def span(self):
        return (
            min(segment.useable_start_time for segment in self),
            max(segment.useable_stop_time for segment in self)
        )



from lxml.etree import SubElement

from oem.tools import parse_utc, parse_str, format_epoch
from oem.base import KeyValueSection, HeaderField


class HeaderSection(KeyValueSection):
    """OEM header section.

    Container for a single OEM header section.

    Examples:
        This class behaves similar to a dict allowing membership checks,
        iteration over keys, and value set/get.

        >>> "CCSDS_OEM_VERS" in header:
        True

        >>> keys = [key for key in header]

        >>> metadata["ORIGINATOR"] = 'ORIG_NAME'

        >>> metadata["ORIGINATOR"]
        'ORIG_NAME'
    """

    _field_spec = {
        "CCSDS_OEM_VERS": HeaderField(parse_str, str, required=True),
        "CREATION_DATE": HeaderField(parse_utc, format_epoch, required=True),
        "ORIGINATOR": HeaderField(parse_str, str, required=True)
    }

    def __init__(self, fields):
        self._parse_fields(fields)

    def __eq__(self, other):
        return (
            self._fields.keys() == other._fields.keys() and
            all(self[key] == other[key] for key in self)
        )

    def __repr__(self):
        return f"HeaderSection(v{self.version})"

    @classmethod
    def _from_raw_data(cls, segment):
        return cls(segment)

    def _to_string(self):
        lines = f"CCSDS_OEM_VERS = {self.version}\n"
        lines += "\n".join([
            entry
            for entry in self._format_fields()
            if "CCSDS_OEM_VERS" not in entry
        ])
        return lines + "\n"

    def _to_xml(self, parent):
        for key, value in self._fields.items():
            if key != "CCSDS_OEM_VERS":
                SubElement(parent, key).text = value

    def copy(self):
        """Create an independent copy of this instance."""
        return HeaderSection(self._fields.copy())

    @property
    def version(self):
        return self["CCSDS_OEM_VERS"]



import pytest

from astropy.time import Time, TimeDelta
from astropy import units
from oem.tle import tle_to_oem


SAMPLE_TLE = (
    '1 25544U 98067A   19343.69339541  .00001764  00000-0  38792-4 0  9991',
    '2 25544  51.6439 211.2001 0007417  17.6667  85.6398 15.50103472202482'
)


@pytest.mark.parametrize("frame", ("TEME", "ICRF"))
def test_sample(frame):
    start_epoch = Time("2019-12-09T20:42:09.000", scale="utc")
    stop_epoch = start_epoch + TimeDelta(1*units.day)
    oem = tle_to_oem(SAMPLE_TLE, start_epoch, stop_epoch, 3600, frame=frame)
    assert len(oem._segments) == 1


@pytest.mark.parametrize("frame", ("TEME", "ICRF"))
def test_convert_and_compare(frame):
    start_epoch = Time("2019-12-09T20:42:09.000", scale="utc")
    stop_epoch = start_epoch + TimeDelta(1*units.day)
    origin = tle_to_oem(SAMPLE_TLE, start_epoch, stop_epoch, 600, frame=frame)
    target = tle_to_oem(SAMPLE_TLE, *origin.span, 600, frame=frame)
    compare = target - origin
    assert not compare.is_empty
    for compare in compare.steps(3600):
        assert compare.range == 0 and compare.range_rate == 0


def test_bad_frame():
    start_epoch = Time("2019-12-09T20:42:09.000", scale="utc")
    stop_epoch = start_epoch + TimeDelta(1*units.day)
    with pytest.raises(ValueError):
        tle_to_oem(SAMPLE_TLE, start_epoch, stop_epoch, 3600, frame="aBcDe")


def test_bad_tle():
    start_epoch = Time("2019-12-09T20:42:09.000", scale="utc")
    stop_epoch = start_epoch + TimeDelta(1*units.day)
    with pytest.raises(ValueError):
        tle_to_oem(["", ""], start_epoch, stop_epoch, 3600)



from enum import Enum
import re
from defusedxml.ElementTree import parse


Section = Enum('Section', ['HEADER', 'META', 'DATA', 'COVARIANCE'])


HS = r"(?:[ \t]|$)+"
"""Arbitrary Horizontal spacing or EOL"""
KEY_VAL = f"([A-Z_]+){HS}={HS}(.+)"
"""Key-value pair"""

COV_XML_KEYS = (
    'CX_X',
    'CY_X',
    'CY_Y',
    'CZ_X',
    'CZ_Y',
    'CZ_Z',
    'CX_DOT_X',
    'CX_DOT_Y',
    'CX_DOT_Z',
    'CX_DOT_X_DOT',
    'CY_DOT_X',
    'CY_DOT_Y',
    'CY_DOT_Z',
    'CY_DOT_X_DOT',
    'CY_DOT_Y_DOT',
    'CZ_DOT_X',
    'CZ_DOT_Y',
    'CZ_DOT_Z',
    'CZ_DOT_X_DOT',
    'CZ_DOT_Y_DOT',
    'CZ_DOT_Z_DOT',
)


def err(line_number, message):
    raise ValueError(f'Error on line {line_number + 2}: {message}')


def parse_kvn_oem(ephem_file):
    section = Section.HEADER
    header, segments = {}, []
    covdata = None
    data_length = None

    match = re.match(KEY_VAL, ephem_file.readline())
    if match:
        header[match.group(1)] = match.group(2)
    if 'CCSDS_OEM_VERS' not in header:
        err('OEM file must start with "CCSDS_OEM_VERS" keyword.')

    for idx, line in enumerate(ephem_file):
        line = line.strip()
        if line == '' or line.startswith('COMMENT'):
            continue  # TODO: Preserve comments

        elif section == Section.DATA:
            if line == 'META_START':
                section = Section.META
                segments.append({'header': {}, 'data': [], 'cov': []})
                continue
            elif line == 'COVARIANCE_START':
                section = Section.COVARIANCE
                continue

            date, *values = line.split()
            if data_length is None:
                data_length = len(values)
                if data_length not in (6, 9):
                    err(idx, 'Malformed data entry')
            elif len(values) != data_length:
                err(idx, 'Data contains mix of data lengths.')
            try:
                values = tuple(float(entry) for entry in values)
            except Exception:
                err(idx, 'Malformed data entry')
            segments[-1]['data'].append((date, *values))

        elif section == Section.COVARIANCE:
            if line == 'COVARIANCE_STOP':
                covdata = None
                continue
            elif line == 'META_START':
                section = Section.META
                segments.append({'header': {}, 'data': [], 'cov': []})
                continue

            if not covdata:
                covdata = {
                    'frame': segments[-1]['header']['REF_FRAME'],
                    'data': (),
                }
                in_header = True

            if '=' in line:
                if not in_header:
                    err(idx, 'Malformed covariance')
                match = re.match(KEY_VAL, line)
                if match:
                    # TODO: Prevent repeats
                    if match.group(1) == 'EPOCH':
                        covdata['epoch'] = match.group(2)
                    elif match.group(1) == 'COV_REF_FRAME':
                        covdata['frame'] = match.group(2)
                    else:
                        err(idx, 'Invalid covariance header')
                else:
                    err(idx, 'Invalid covariance header')
            else:
                if in_header:
                    in_header = False
                    cov_data_line = 1
                    covdata['data'] = covdata['data'] + (float(line),)
                else:
                    cov_data_line += 1
                    raw_values = line.split()
                    if len(raw_values) != cov_data_line:
                        err(idx, 'Malformed covariance shape')
                    covdata['data'] = (
                        covdata['data'] +
                        tuple(float(entry) for entry in raw_values)
                    )

                    if cov_data_line == 6:
                        segments[-1]['cov'].append((
                            covdata['epoch'],
                            covdata['frame'],
                            *covdata['data'],
                        ))
                        covdata = None

        elif section == Section.HEADER:
            if line == 'META_START':
                section = Section.META
                segments.append({'header': {}, 'data': [], 'cov': []})
                continue

            match = re.match(KEY_VAL, line)
            if match:
                if match.group(1) in header:
                    err(idx, f'Duplicate header: {match.group(1)}')
                header[match.group(1)] = match.group(2)
            else:
                err(idx, 'Invalid header entry')

        elif section == Section.META:
            if line == 'META_STOP':
                section = Section.DATA
                data_length = None
                continue

            match = re.match(KEY_VAL, line)
            if match:
                if match.group(1) in header:
                    err(idx, f'Duplicate entry: {match.group(1)}')
                segments[-1]['header'][match.group(1)] = match.group(2)
            else:
                err(idx, 'Invalid meta entry')

    return header, segments


def parse_xml_oem(ephem_file):
    parts = parse(ephem_file).getroot()

    header = {
        entry.tag.rpartition('}')[-1]: entry.text
        for entry in parts[0]
        if entry.tag.rpartition('}')[-1] != "COMMENT"
    }
    header['CCSDS_OEM_VERS'] = parts.attrib['version']

    segments = []
    for raw_segment in parts[1]:
        raw_metadata, raw_data = raw_segment

        segment = {}
        segment['header'] = {
            entry.tag.rpartition('}')[-1]: entry.text
            for entry in raw_metadata
            if entry.tag.rpartition('}')[-1] != "COMMENT"
        }

        keys = ('X', 'Y', 'Z', 'X_DOT', 'Y_DOT', 'Z_DOT')
        try:
            if raw_data.find('stateVector').find('X_DDOT') is not None:
                keys = keys + ('X_DDOT', 'Y_DDOT', 'Z_DDOT')
            segment['data'] = tuple(
                (entry.find('EPOCH').text,)
                + tuple(float(entry.find(key).text) for key in keys)
                for entry in raw_data
                if entry.tag.rpartition('}')[-1] == "stateVector"
            )
        except Exception:
            raise ValueError('Malformed data section')

        ref_frame = raw_metadata.find('REF_FRAME')
        try:
            segment['cov'] = tuple(
                (
                    entry.find('EPOCH').text,
                    (entry.find('COV_REF_FRAME') or ref_frame).text,
                    *(float(entry.find(key).text) for key in COV_XML_KEYS)
                )
                for entry in raw_data
                if entry.tag.rpartition('}')[-1] == "covarianceMatrix"
            )
        except Exception:
            raise ValueError('Malformed covariance section')

        segments.append(segment)

    return header, segments



import numpy as np


def lagrange(x, y):
    """Create a Lagrange interpolation polynomial.

    Create a Lagrange interpolation polynomial of order N-1 where N is the
    number of (x, y) coordinates provided.

    Args:
        x (ndarray): Interpolation point x values, length N.
        y (ndarray): Interpolation point y values, length N.

    Returns:
        poly (poly1d): Polynomial object called with poly(x).
    """
    order = x.size - 1
    A = np.power(
        np.tile(x, (order + 1, 1)).T,
        np.arange(order + 1)
    )
    a = np.linalg.solve(A, y)
    return np.poly1d(a[::-1])


def hermite(x, y, dy):
    """Create a Hermite interpolation polynomial.

    Create a Hermite interpolation polynomial of order N-1 where N is the
    number of (x, y, dy) entries provided.

    Args:
        x (ndarray): Interpolation point x values, length N.
        y (ndarray): Interpolation point y values, length N.
        dy (ndarray): Interpolation point dy/dx values, length N.

    Returns:
        poly (poly1d): Polynomial object called with poly(x).
    """
    order = 2*x.size - 1
    c = np.tile(x, (order + 1, 1)).T
    Au = np.power(
        c,
        np.arange(order + 1)
    )
    Al = np.multiply(
        np.power(
            c,
            np.hstack((
                [1],
                np.arange(order)
            ))
        ),
        np.tile(np.hstack(([0, 1], np.arange(2, order+1))), (x.size, 1))
    )
    A = np.vstack((Au, Al))
    b = np.hstack((y, dy))
    a = np.linalg.solve(A, b)
    return np.poly1d(a[::-1])


class Interpolator(object):

    def __init__(self, states):
        self._reference_epoch = states[0][0]
        self._setup(states)

    def __call__(self, epoch):
        t = (epoch - self.reference_epoch).sec
        raw_state = np.array([poly(t) for poly in self._state_polynomials])
        position = raw_state[:3]
        velocity = raw_state[3:6]
        if len(raw_state) == 9:
            acceleration = raw_state[6:]
        else:
            acceleration = None
        return position, velocity, acceleration

    def _elapsed_times(self, states):
        reference = self.reference_epoch
        return np.array(tuple((epoch - reference).sec for epoch in states[0]))

    @property
    def reference_epoch(self):
        return self._reference_epoch


class LagrangeStateInterpolator(Interpolator):

    @classmethod
    def _samples_required(cls, order):
        count = order + 1
        if count % 1 != 0:
            raise ValueError("Unachievable order: {order}")
        else:
            return int(count)

    def _setup(self, states):
        t = self._elapsed_times(states)
        state_vectors = np.column_stack(states[1:])
        self._state_polynomials = [
            lagrange(t, state_vectors[:, idx])
            for idx in range(state_vectors.shape[1])
        ]


class HermiteStateInterpolator(Interpolator):

    @classmethod
    def _samples_required(cls, order):
        count = (order + 1)/2
        if count % 1 != 0:
            raise ValueError("Unachievable order: {order}")
        else:
            return int(count)

    def _setup(self, states):
        t = self._elapsed_times(states)
        state_vectors = np.column_stack(states[1:])
        self._state_polynomials = [
            hermite(t, state_vectors[:, idx], state_vectors[:, idx+3])
            for idx in range(3)
        ]

        if state_vectors.shape[1] == 9:
            self._state_polynomials += [
                hermite(t, state_vectors[:, idx+3], state_vectors[:, idx+6])
                for idx in range(3)
            ]
            self._state_polynomials += [
                entry.deriv() for entry in self._state_polynomials[3:]
            ]

        else:
            self._state_polynomials += [
                entry.deriv() for entry in self._state_polynomials
            ]


class EphemerisInterpolator(object):

    method_map = {
        "lagrange": LagrangeStateInterpolator,
        "hermite": HermiteStateInterpolator
    }

    def __init__(self, states, method, order):
        self.base_interpolator = self.method_map[method.lower()]
        self._states = states
        self._order = order
        self._populate_interpolator_nodes(states[0], order)

    def __call__(self, epoch):
        interpolator = self._get_best_interpolator(epoch)
        return interpolator(epoch)

    def _populate_interpolator_nodes(self, epochs, order):
        samples = self.base_interpolator._samples_required(order)
        elapsed_times = np.array([
            (entry - self.reference_epoch).sec for entry in epochs]
        )
        self._nodes = np.array([
            np.mean(elapsed_times[idx:(idx + samples)])
            for idx in range(len(elapsed_times) - samples + 1)
        ])

    def _get_best_interpolator(self, epoch):
        elapsed_time = (epoch - self.reference_epoch).sec
        best_idx = np.argmin(np.abs(self._nodes - elapsed_time))
        samples = self.base_interpolator._samples_required(self.order)
        return self.base_interpolator(
            tuple(entry[best_idx:best_idx+samples] for entry in self._states)
        )

    @property
    def reference_epoch(self):
        return self._states[0][0]

    @property
    def order(self):
        return self._order



import pytest
import numpy as np

from astropy.time import Time
from pathlib import Path
from oem import OrbitEphemerisMessage
from oem.compare import StateCompare
from oem.components import State


SAMPLE_DIR = Path(__file__).parent / "samples"


def test_state_self_difference():
    state = State(Time.now(), "ICRF", "EARTH", [1, 0, 0], [0, 1, 0], [0, 0, 1])
    compare = state - state
    assert compare.range == 0
    assert compare.range_rate == 0
    assert all(compare.position == 0)
    assert all(compare.velocity == 0)
    assert all(compare.position_ric == 0)
    assert all(compare.velocity_ric == 0)


def test_state_compare_frame_mismatch():
    epoch = Time.now()
    origin = State(epoch, "ICRF", "EARTH", [1, 0, 0], [0, 1, 0])
    target1 = State(epoch, "GRC", "EARTH", [1, 0, 0], [0, 1, 0])
    target2 = State(epoch, "ICRF", "MARS", [1, 0, 0], [0, 1, 0])
    with pytest.raises(ValueError):
        StateCompare(origin, target1)
    with pytest.raises(ValueError):
        StateCompare(origin, target2)


def test_state_compare_noninertial():
    state = State(Time.now(), "GRC", "EARTH", [1, 0, 0], [0, 1, 0])
    with pytest.raises(NotImplementedError):
        StateCompare(state, state).velocity


def test_state_compare_nonstandard():
    state = State(Time.now(), "ABCD", "EARTH", [1, 0, 0], [0, 1, 0])
    with pytest.warns(UserWarning):
        StateCompare(state, state)


def test_state_compare_epoch_mismatch():
    origin = State(Time.now(), "ICRF", "EARTH", [1, 0, 0], [0, 1, 0])
    target = State(Time.now(), "ICRF", "EARTH", [1, 0, 0], [0, 1, 0])
    with pytest.raises(ValueError):
        StateCompare(origin, target)


def test_segment_self_compare():
    test_file_path = SAMPLE_DIR / "real" / "GEO_20s.oem"
    segment = OrbitEphemerisMessage.open(test_file_path).segments[0]
    compare = segment - segment
    assert not compare.is_empty
    for state_compare in compare.steps(600):
        assert state_compare.range == 0 and state_compare.range_rate == 0


def test_segment_compare_mismatch():
    test_file_path = SAMPLE_DIR / "real" / "GEO_20s.oem"
    segment1 = OrbitEphemerisMessage.open(test_file_path).segments[0]
    segment2 = segment1.copy()
    _ = segment1 - segment2
    segment2.metadata["CENTER_NAME"] = "MARS"
    with pytest.raises(ValueError):
        _ = segment1 - segment2


def test_ephemeris_self_compare():
    test_file_path = SAMPLE_DIR / "real" / "GEO_20s.oem"
    oem = OrbitEphemerisMessage.open(test_file_path)
    compare = oem - oem
    assert not compare.is_empty
    for state_compare in compare.steps(600):
        assert state_compare.range == 0 and state_compare.range_rate == 0
        np.testing.assert_almost_equal(state_compare.position_ric, 0)
        np.testing.assert_almost_equal(state_compare.velocity_ric, 0)


def test_real_reference_ric():
    test_origin_path = SAMPLE_DIR / "real" / "CompareExample1.oem"
    test_target_path = SAMPLE_DIR / "real" / "CompareExample2.oem"
    origin = OrbitEphemerisMessage.open(test_origin_path)
    target = OrbitEphemerisMessage.open(test_target_path)
    compare = target - origin
    assert not compare.is_empty
    for state_compare in compare.steps(600):
        np.testing.assert_almost_equal(state_compare.range, 1.165554784013)
        np.testing.assert_almost_equal(
            state_compare.position_ric,
            np.array([-0.000101713843, -1.165554779575, 0.0]),
            decimal=6
        )
        np.testing.assert_almost_equal(state_compare.velocity_ric, 0)



from lxml.etree import SubElement
from itertools import chain

from oem import CURRENT_VERSION
from oem.base import ConstraintSpecification, Constraint
from oem.tools import (
    require, time_range, epoch_span_contains, _bulk_parse_epochs, format_float,
    format_epoch
)
from oem.parsers import COV_XML_KEYS
from oem.components.metadata import MetaDataSection
from oem.components.types import State, Covariance
from oem.interp import EphemerisInterpolator
from oem.compare import SegmentCompare


class ConstrainEphemerisSegmentCovariance(Constraint):
    '''Apply constraints to ephemeris segment covariance sections'''

    versions = ["1.0"]

    def func(self, ephemeris_segment):
        require(
            ephemeris_segment._covariance_data is None,
            "Covariance data block not supported in OEM v1.0"
        )


class ConstrainEphemerisSegmentStateVectors(Constraint):
    '''Apply constraints to ephemeris segment state vectors'''

    versions = ["1.0"]

    def func(self, ephemeris_segment):
        require(
            not ephemeris_segment.has_accel,
            "Acceleration is not supported in v1.0 OEMs",
        )


def _process_states(metadata, raw_data_rows):
    if len(raw_data_rows) == 0:
        raise ValueError('Empty data section.')

    raw_data_columns = tuple(zip(*raw_data_rows))

    epochs = raw_data_columns[0]
    sorted = all(epochs[idx] < epochs[idx+1] for idx in range(len(epochs) - 1))
    require(sorted, "States in data section are not ordered by epoch")

    epochs = tuple(_bulk_parse_epochs(raw_data_columns[0], metadata))
    return (epochs, *raw_data_columns[1:])


def _process_covariances(metadata, raw_data_rows):
    if len(raw_data_rows) == 0:
        raise ValueError('Empty covariance section.')
    raw_data_columns = tuple(zip(*raw_data_rows))
    epochs = tuple(_bulk_parse_epochs(raw_data_columns[0], metadata))
    frames = raw_data_columns[1]
    return (epochs, frames, *raw_data_columns[2:])


class EphemerisSegment(object):
    """OEM ephemeris segment.

    Container for a single OEM ephemeris segment.
    """

    _constraint_spec = ConstraintSpecification(
        ConstrainEphemerisSegmentCovariance,
        ConstrainEphemerisSegmentStateVectors,
    )

    def __init__(self, metadata, state_data, covariance_data=None,
                 version=CURRENT_VERSION):
        self.version = version
        self.metadata = metadata
        self._state_data = state_data
        self._covariance_data = covariance_data
        self._constraint_spec.apply(self)
        self._interpolator = None

    def __call__(self, epoch):
        if epoch not in self:
            raise ValueError(f"Epoch {epoch} not contained in segment.")
        if not self._interpolator:
            self._init_interpolator()
        position, velocity, acceleration = self._interpolator(epoch)
        return State(
            epoch,
            self.metadata["REF_FRAME"],
            self.metadata["CENTER_NAME"],
            position,
            velocity,
            acceleration=acceleration,
            version=self.version,
        )

    def __contains__(self, epoch):
        return epoch_span_contains(self.span, epoch)

    def __iter__(self):
        return self.states

    def __eq__(self, other):
        return (
            self.version == other.version and
            self.metadata == other.metadata and
            self._state_data == other._state_data and
            self._covariance_data == other._covariance_data
        )

    def __sub__(self, other):
        return SegmentCompare(other, self)

    def __repr__(self):
        start = str(self.useable_start_time)
        stop = str(self.useable_stop_time)
        return f"EphemerisSegment({start}, {stop})"

    @classmethod
    def _from_raw_data(cls, segment, version):
        metadata = MetaDataSection._from_raw_data(segment['header'], version)

        try:
            state_data = _process_states(metadata, segment['data'])
        except Exception:
            raise ValueError('Malformed data section.')

        if segment.get('cov'):
            try:
                cov_data = _process_covariances(metadata, segment['cov'])
            except Exception:
                raise ValueError('Malformed covariance section.')
        else:
            cov_data = None

        return cls(metadata, state_data, cov_data, version=version)

    def _to_string(self):
        lines = self.metadata._to_string() + "\n"
        for epoch, *state in zip(*self._state_data):
            lines += f'{format_epoch(epoch)} '
            lines += ' '.join(format_float(entry) for entry in state) + '\n'
        lines += '\n'

        if self._covariance_data:
            lines += 'COVARIANCE_START\n'
            for epoch, frame, *cov in zip(*self._covariance_data):
                lines += f'EPOCH = {format_epoch(epoch)}\n'
                if frame != self.metadata['REF_FRAME']:
                    lines += f'COV_REF_FRAME = {frame}\n'
                lines += f'{format_float(cov[0])}\n'
                lines += ' '.join(format_float(c) for c in cov[1:3]) + '\n'
                lines += ' '.join(format_float(c) for c in cov[3:6]) + '\n'
                lines += ' '.join(format_float(c) for c in cov[6:10]) + '\n'
                lines += ' '.join(format_float(c) for c in cov[10:15]) + '\n'
                lines += ' '.join(format_float(c) for c in cov[15:]) + '\n'
            lines += 'COVARIANCE_STOP\n'
            lines += '\n'

        return lines

    def _to_xml(self, parent):
        self.metadata._to_xml(SubElement(parent, "metadata"))
        data = SubElement(parent, "data")
        fields = ('X', 'Y', 'Z', 'X_DOT', 'Y_DOT', 'Z_DOT')
        for epoch, *state in zip(*self._state_data):
            vector = SubElement(data, 'stateVector')
            SubElement(vector, "EPOCH").text = format_epoch(epoch)
            for idx, field in enumerate(fields):
                SubElement(vector, field).text = format_float(state[idx])
            if len(state) == 9:
                for idx, field in enumerate(('X_DDOT', 'Y_DDOT', 'Z_DDOT')):
                    SubElement(vector, field).text = format_float(state[idx+6])

        if self._covariance_data:
            for epoch, frame, *cov in zip(*self._covariance_data):
                covdata = SubElement(data, 'covarianceMatrix')
                SubElement(covdata, "EPOCH").text = format_epoch(epoch)
                if frame != self.metadata['REF_FRAME']:
                    SubElement(covdata, "COV_REF_FRAME").text = frame
                for idx, key in enumerate(COV_XML_KEYS):
                    SubElement(covdata, key).text = format_float(cov[idx])

    def _init_interpolator(self):
        if "INTERPOLATION" in self.metadata:
            method = self.metadata["INTERPOLATION"]
            order = self.metadata["INTERPOLATION_DEGREE"]
        else:
            method = "LAGRANGE"
            order = 5
        self._interpolator = EphemerisInterpolator(
            self._state_data,
            method,
            order
        )

    def copy(self):
        """Create an independent copy of this instance."""
        return EphemerisSegment(
            self.metadata.copy(),
            self._state_data,
            self._covariance_data if self.has_covariance else None,
            version=self.version
        )

    def steps(self, step_size):
        """Sample Segment at equal time intervals.

        This method returns a generator producing states at equal time
        intervals spanning the useable duration of the parent EphemerisSegment.

        Args:
            step_size (float): Sample step size in seconds.

        Yields:
            State: Sampled state.

        Examples:
            Sample states in each segment of an OEM at 60-second intervals:

            >>> for segment in oem:
            ...    for state in segment.steps(60):
            ...        pass
        """
        for epoch in time_range(self.useable_start_time,
                                self.useable_stop_time,
                                step_size):
            yield self(epoch)

    def resample(self, step_size, in_place=False):
        """Resample ephemeris data.

        Replaces the existing ephemeris state data in this EphemerisSegment
        with a new list of states sampled at the desired sampling interval.

        Args:
            step_size (float): Sample step size in seconds.
            in_place (bool, optional): Toggle in-place resampling. Default
                is False.

        Returns:
            EphemerisSegment: Resampled EphemerisSegment. Output is
                an indepdent instance if in_place is True.
        """
        if not self._interpolator:
            self._init_interpolator()

        epochs = time_range(
            self.useable_start_time, self.useable_stop_time, step_size
        )

        if in_place:
            states = (
                (
                    epoch, *chain.from_iterable(
                        self._interpolator(epoch)[:2 + self.has_accel]
                    )
                ) for epoch in epochs
            )
            self._state_data = tuple(zip(*states))
        else:
            segment = self.copy().resample(step_size, in_place=True)

        return segment if not in_place else self

    @property
    def states(self):
        """Return list of States in this segment."""
        return (
            State._from_raw_data(entry, self.version, self.metadata)
            for entry in zip(*self._state_data)
        )

    @property
    def covariances(self):
        """Return list of Covariances in this segment."""
        if self._covariance_data:
            return (
                Covariance._from_raw_data(entry, self.version)
                for entry in zip(*self._covariance_data)
            )
        else:
            return ()

    @property
    def has_accel(self):
        """Evaluate if segment contains acceleration data."""
        return len(self._state_data) == 10

    @property
    def has_covariance(self):
        """Evaluate if segment contains covariance data."""
        return True if self._covariance_data else False

    @property
    def useable_start_time(self):
        """Return epoch of start of useable state data range"""
        return self.metadata.useable_start_time

    @property
    def useable_stop_time(self):
        """Return epoch of end of useable state data range"""
        return self.metadata.useable_stop_time

    @property
    def span(self):
        return (self.useable_start_time, self.useable_stop_time)



import pytest
import datetime as dt
from pathlib import Path

import numpy as np
from astropy.time import Time, TimeDelta

from oem import OrbitEphemerisMessage
from oem.interp import (
    LagrangeStateInterpolator, HermiteStateInterpolator, EphemerisInterpolator)


THIS_DIR = Path(__file__).parent
SAMPLE_DIR = THIS_DIR / "samples"


def _make_test_states(poly, t_step, count, accel=True):
    """Create state samples for testing.

    Args:
        poly (poly1d): Polynomial describing position history.
        t_step (float): State time step in seconds.
        count (int): Number of points to sample.
        accel (bool, optional): If True, output States will have accelerations.

    Returns:
        states (list): List of State with position, velocity, and acceleration
            following poly, poly.deriv, and poly.deriv.deriv, respectively.
            Epochs are DateTimes starting at the current time and stepping
            by t_step.
    """
    start_epoch = dt.datetime.now()
    epochs = [
        Time(start_epoch + dt.timedelta(seconds=t_step*idx))
        for idx in range(count)
    ]
    positions = [poly([t_step*idx]*3) for idx in range(count)]
    velocities = [poly.deriv()([t_step*idx]*3) for idx in range(count)]
    if accel:
        accelerations = [
            poly.deriv().deriv()([t_step*idx]*3)
            for idx in range(count)
        ]
        return (
            epochs, *zip(*positions), *zip(*velocities), *zip(*accelerations)
        )
    else:
        return (epochs, *zip(*positions), *zip(*velocities))


@pytest.mark.parametrize("has_accel", (True, False))
@pytest.mark.parametrize(
    "Interpolator, samples",
    ((LagrangeStateInterpolator, 8), (HermiteStateInterpolator, 4))
)
def test_interpolators(Interpolator, samples, has_accel):
    position = np.poly1d([.1, .1, .1])
    velocity = position.deriv()
    acceleration = velocity.deriv()
    time_step = 60

    states = _make_test_states(position, time_step, samples, accel=has_accel)
    interpolator = Interpolator(states)

    for elapsed in np.arange(0, (samples-1)*time_step, 1):
        test_epoch = states[0][0] + TimeDelta(elapsed, format="sec")
        predict_pos, predict_vel, predict_accel = interpolator(test_epoch)
        np.testing.assert_almost_equal(predict_pos, position(elapsed))
        np.testing.assert_almost_equal(predict_vel, velocity(elapsed))
        if has_accel:
            np.testing.assert_almost_equal(predict_accel, acceleration(elapsed))


@pytest.mark.parametrize("has_accel", (True, False))
@pytest.mark.parametrize("method, order", (("LAGRANGE", 8), ("HERMITE", 9)))
def test_ephemeris_interpolator(method, order, has_accel):
    position = np.poly1d([.1, .1, .1])
    velocity = position.deriv()
    acceleration = velocity.deriv()
    time_step = 30
    samples = 25

    states = _make_test_states(position, time_step, samples, accel=has_accel)
    interpolator = EphemerisInterpolator(states, method, order)

    for elapsed in np.arange(0, (samples-1)*time_step, 5):
        test_epoch = states[0][0] + TimeDelta(elapsed, format="sec")
        predict_pos, predict_vel, predict_accel = interpolator(test_epoch)
        np.testing.assert_almost_equal(predict_pos, position(elapsed), 6)
        np.testing.assert_almost_equal(predict_vel, velocity(elapsed), 6)
        if has_accel:
            np.testing.assert_almost_equal(
                predict_accel, acceleration(elapsed), 6
            )


@pytest.mark.parametrize(
    "coarse_file, fine_file",
    (
        ("GEO_60s.oem", "GEO_20s.oem"),
        ("MEO_60s.oem", "MEO_20s.oem"),
        ("LEO_60s.oem", "LEO_10s.oem")
    )
)
def test_ephemeris_accuracy(coarse_file, fine_file):
    fine_sample = SAMPLE_DIR / "real" / fine_file
    coarse_sample = SAMPLE_DIR / "real" / coarse_file
    fine_oem = OrbitEphemerisMessage.open(fine_sample)
    coarse_oem = OrbitEphemerisMessage.open(coarse_sample)

    for state in fine_oem.states:
        predict = coarse_oem(state.epoch)
        np.testing.assert_almost_equal(predict.position, state.position, 6)
        np.testing.assert_almost_equal(predict.velocity, state.velocity, 6)
        if state.has_accel:
            np.testing.assert_almost_equal(
                predict.acceleration,
                state.acceleration,
                6
            )


@pytest.mark.parametrize(
    "input_file", ("GEO_20s.oem", "MEO_20s.oem", "LEO_10s.oem")
)
def test_ephemeris_stepping(input_file):
    sample_file = SAMPLE_DIR / "real" / input_file
    oem = OrbitEphemerisMessage.open(sample_file)

    for state in oem.steps(601):
        assert state.epoch in oem

    for segment in oem:
        for state in segment.steps(601):
            assert state.epoch in oem

    out_of_bounds_epoch = oem.span[0] - TimeDelta(1)
    with pytest.raises(ValueError):
        oem(out_of_bounds_epoch)
    for segment in oem:
        with pytest.raises(ValueError):
            segment(out_of_bounds_epoch)


@pytest.mark.parametrize(
    "input_file", ("GEO_20s.oem", "MEO_20s.oem", "LEO_10s.oem")
)
def test_ephemeris_resample(input_file):
    sample_file = SAMPLE_DIR / "real" / input_file
    step_size = 600
    oem = OrbitEphemerisMessage.open(sample_file)
    new_oem = oem.resample(step_size)

    for idx in range(1, len(new_oem.states)):
        assert np.isclose(
            (new_oem.states[idx].epoch - new_oem.states[idx-1].epoch).sec,
            step_size
        )



import numpy as np
import warnings

from oem.tools import epoch_span_overlap, epoch_span_contains, time_range


REFERENCE_FRAMES = {
    "inertial": ["EME2000", "GCRF", "ICRF", "MCI", "TEME", "TOD"],
    "rotating": ["GRC", "ITRF2000", "ITRF-93", "ITRF-97", "TDR"]
}


class EphemerisCompare(object):
    """Comparison of two OrbitEphemerisMessage.

    Only overlapping segments with identical reference frame and central bodies
    will be compared. All comparisons are calculated in the segment reference
    frame. Rotating reference frames are not supported for velocity-based or
    RIC comparisons.

    Attributes:
        is_empty (bool): Flag indicating overlap between compared ephemerides.
            Set to True if there is no overlap.
        segments (list): List of SegmentCompare for matching EphemerisSegment
            with overlapping spans.

    Examples:
        A comparison of two ephemerides is simply achieved through either a
        direct call to EphemerisCompare or the subtraction interface on
        the OrbitEphemerisMessage class. In general, the subtraction interface
        is preferred.

        >>> ephemeris1 = OrbitEphemerisMessage.open(file_path1)
        >>> ephemeris2 = OrbitEphemerisMessage.open(file_path2)
        >>> compare = ephemeris2 - ephemeris1

        The EphemerisCompare object supports most of the same basic interfaces
        as OrbitEphemerisMessage. To evaluate the at a particular epoch:

        >>> compare(epoch)

        To iterate through the compare at a fixed interval, use the `.steps`
        method:

        >>> for state_compare in compare.steps(60):
        ...     # Operate on StateCompare
        ...     pass

        For multi-segment ephemerides, EphemerisCompare is iterable:

        >>> for segment_compare in compare:
        ...     for state_compare in segment_compare:
        ...     pass
    """

    def __init__(self, origin, target):
        """Create an EphemerisCompare.

        Args:
            origin (OrbitEphemerisMessage): Ephemeris at the origin of the
                compare frame.
            target (OrbitEphemerisMessage): Ephemeris compared against origin.
        """
        segments = []
        for origin_segment in origin:
            for target_segment in target:
                try:
                    segments.append(target_segment - origin_segment)
                except ValueError:
                    continue
        self._segments = [entry for entry in segments if not entry.is_empty]

    def __call__(self, epoch):
        for segment in self:
            if epoch in segment:
                return segment(epoch)
        else:
            raise ValueError(
                f"Epoch {epoch} not contained in EphemerisCompare."
            )

    def __iter__(self):
        return iter(self._segments)

    def __contains__(self, epoch):
        return any(epoch in segment for segment in self._segments)

    def __repr__(self):
        return f"EphemerisCompare(segments: {len(self.segments)})"

    def steps(self, step_size):
        """Sample EphemerisCompare at equal time intervals.

        This method returns a generator producing state compares at equal time
        intervals spanning the useable duration of the parent EphemerisCompare.

        Args:
            step_size (float): Sample step size in seconds.

        Yields:
            state_compare: Sampled StateCompare.
        """
        for segment in self:
            for state in segment.steps(step_size):
                yield state

    @property
    def is_empty(self):
        return len(self._segments) == 0

    @property
    def segments(self):
        return self._segments


class SegmentCompare(object):
    """Comparison of two EphemerisSegment.

    Input segments must have identical reference frames and central bodies.
    All comparisons are calculated in the input segment reference frame.

    Rotating reference frames are not supported for velocity-based or
    RIC comparisons.

    Attributes:
        is_empty (bool): Flag indicating overlap between compared segments. Set
            to True if there is no overlap.
    """

    def __init__(self, origin, target):
        """Create a SegmentCompare.

        Args:
            origin (EphemerisSegment): Segment at the origin of the
                compare frame.
            target (EphemerisSegment): Segment to compare against the
                origin state.
        """
        if (origin.metadata["REF_FRAME"] == target.metadata["REF_FRAME"]
                and origin.metadata["CENTER_NAME"]
                == target.metadata["CENTER_NAME"]):
            self._span = epoch_span_overlap(origin.span, target.span)
            self._origin = origin
            self._target = target
        else:
            raise ValueError(
                "Incompatible states: frame or central body mismatch."
            )

    def __contains__(self, epoch):
        return (
            self._span is not None and epoch_span_contains(self._span, epoch)
        )

    def __call__(self, epoch):
        if epoch not in self:
            raise ValueError(f"Epoch {epoch} not contained in SegmentCompare.")
        return self._target(epoch) - self._origin(epoch)

    def __repr__(self):
        return f"SegmentCompare({str(self._span[0])}, {str(self._span[1])})"

    def steps(self, step_size):
        """Sample SegmentCompare at equal time intervals.

        This method returns a generator producing state compares at equal time
        intervals spanning the useable duration of the parent SegmentCompare.

        Args:
            step_size (float): Sample step size in seconds.

        Yields:
            state_compare: Sampled StateCompare.
        """
        for epoch in time_range(*self._span, step_size):
            yield self(epoch)

    @property
    def is_empty(self):
        return self._span is None


class StateCompare(object):
    """Comparison of two Cartesian states.

    Input states must have identical epochs, reference frames, and central
    bodies. All comparisons are calculated in the input state reference frame.

    Rotating reference frames are not supported for velocity-based or
    RIC comparisons.

    Attributes:
        epoch (Time): Epoch of the state compare.
        range (float): Absolute distance between the two states.
        range_rate (float): Absolute velocity between the two states.
        position (ndarray): Relative position vector in the input frame.
        velocity (ndarray): Relative velocity vector in the input frame.
        position_ric (ndarray): Relative position vector in the RIC frame.
        velocity_ric (ndarray): Relative velocity vector in the RIC frame.

    Examples:
        To compare two states, `origin` and `target`, either call the
        StateCompare initializer directly

        >>> compare = StateCompare(origin, target)

        or simply difference the two states

        >>> compare = origin - target
    """

    def __init__(self, origin, target):
        """Create a StateCompare.

        Args:
            origin (State): State at the origin of the compare frame.
            target (State): State to compare against the origin state.

        Raises:
            ValueError: Incompatible states: epoch, frame, or central
                body mismatch.
        """
        if (origin.epoch == target.epoch
                and origin.frame == target.frame
                and origin.center == target.center):
            self._origin = origin
            self._target = target
            if self._origin.frame.upper() in REFERENCE_FRAMES["inertial"]:
                self._inertial = True
            elif self._origin.frame.upper() in REFERENCE_FRAMES["rotating"]:
                self._inertial = False
            else:
                warnings.warn(
                    f"Nonstandard frame: '{self._origin.frame}'. "
                    "Assuming intertial. Override with ._inertial=False",
                    UserWarning
                )
                self._inertial = True
        else:
            raise ValueError(
                "Incompatible states: epoch, frame, or central body mismatch."
            )

    def __repr__(self):
        return f"StateCompare({str(self.epoch)})"

    def _require_inertial(self):
        if not self._inertial:
            raise NotImplementedError(
                "Velocity compares not supported for non-inertial frames. "
                "To override, set ._inertial=True."
            )

    def _to_ric(self, vector):
        self._require_inertial()
        cross_track = np.cross(self._origin.position, self._origin.velocity)
        in_track = np.cross(cross_track, self._origin.position)
        R = np.array([
            self._origin.position/np.linalg.norm(self._origin.position),
            in_track/np.linalg.norm(in_track),
            cross_track/np.linalg.norm(cross_track)
        ])
        return R.dot(vector)

    @property
    def epoch(self):
        return self._origin.epoch

    @property
    def range(self):
        return np.linalg.norm(self._target.position - self._origin.position)

    @property
    def range_rate(self):
        self._require_inertial()
        return np.linalg.norm(self._target.velocity - self._origin.velocity)

    @property
    def position(self):
        return self._target.position - self._origin.position

    @property
    def velocity(self):
        self._require_inertial()
        return self._target.velocity - self._origin.velocity

    @property
    def position_ric(self):
        return self._to_ric(self.position)

    @property
    def velocity_ric(self):
        w = self._to_ric(
            np.cross(self._origin.position, self._origin.velocity)
            / np.linalg.norm(self._origin.position)**2
        )
        return self._to_ric(self.velocity) - np.cross(w, self.position_ric)



from .types import State, Covariance
from .segment import EphemerisSegment
from .header import HeaderSection
from .metadata import MetaDataSection


__all__ = [
    State,
    Covariance,
    EphemerisSegment,
    HeaderSection,
    MetaDataSection,
]
